from argparse import ArgumentParser

from cc_agency.commons.conf import Conf
from cc_agency.commons.db import Mongo

DESCRIPTION = 'Drop MongoDB collections.'


def attach_args(parser):
    parser.add_argument(
        '-c', '--conf-file', action='store', type=str, metavar='CONF_FILE',
        help='CONF_FILE (yaml) as local path.'
    )
    parser.add_argument(
        action='store', type=str, nargs='+', metavar='COLLECTIONS', dest='collections',
        choices=['experiments', 'batches', 'users', 'tokens', 'block_entries', 'callback_tokens'],
        help='Collections to be dropped.'
    )


def main():
    parser = ArgumentParser(description=DESCRIPTION)
    attach_args(parser)
    args = parser.parse_args()

    return run(**args.__dict__)


def run(conf_file, collections):
    conf = Conf(conf_file)
    mongo = Mongo(conf)

    for collection in collections:
        mongo.db[collection].drop()



import os
import jsonschema
from ruamel.yaml import YAML

from cc_agency.commons.schemas.conf import conf_schema

yaml = YAML(typ='safe')


class Conf:
    def __init__(self, conf_file_path):
        if not conf_file_path:
            conf_file_path = os.path.join('~', '.config', 'cc-agency.yml')

        conf_file_path = os.path.expanduser(conf_file_path)

        with open(conf_file_path) as f:
            self.d = yaml.load(f)

        jsonschema.validate(self.d, conf_schema)



# TODO: Incomplete
agent_result_schema = {
    'type': 'object',
    'properties': {
        'state': {'enum': ['succeeded', 'failed']}
    },
    'additionalProperties': True,
    'required': ['state']
}



from flask import jsonify, request
from werkzeug.exceptions import Unauthorized


def auth_routes(app, auth, conf):
    tokens_valid_for_seconds = conf.d['broker']['auth']['tokens_valid_for_seconds']

    @app.route('/token', methods=['GET'])
    def get_token():
        user = auth.verify_user(request.authorization)
        if not user:
            raise Unauthorized()

        if not user['verified_by_credentials']:
            raise Unauthorized()

        token = auth.issue_token(user)
        return jsonify({
            'token': token,
            'validForSeconds': tokens_valid_for_seconds
        })



import sys
from collections import OrderedDict
from argparse import ArgumentParser

from cc_agency.version import VERSION

from cc_agency.tools.create_db_user.main import main as create_db_user_main
from cc_agency.tools.create_broker_user.main import main as create_broker_user_main
from cc_agency.tools.drop_db_collections.main import main as drop_db_collections_main

from cc_agency.tools.create_db_user.main import DESCRIPTION as CREATE_DB_USER_DESCRIPTION
from cc_agency.tools.create_broker_user.main import DESCRIPTION as CREATE_BROKER_USER_DESCRIPTION
from cc_agency.tools.drop_db_collections.main import DESCRIPTION as DROP_DB_COLLECTIONS_DESCRIPTION


SCRIPT_NAME = 'ccagency'

DESCRIPTION = 'CC-Agency Copyright (C) 2018  Christoph Jansen. This software is distributed under the AGPL-3.0 ' \
              'LICENSE and is part of the Curious Containers project (https://curious-containers.github.io/).'

MODES = OrderedDict([
    ('create-db-user', {'main': create_db_user_main, 'description': CREATE_DB_USER_DESCRIPTION}),
    ('create-broker-user', {'main': create_broker_user_main, 'description': CREATE_BROKER_USER_DESCRIPTION}),
    ('drop-db-collections', {'main': drop_db_collections_main, 'description': DROP_DB_COLLECTIONS_DESCRIPTION})
])


def main():
    sys.argv[0] = SCRIPT_NAME

    parser = ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        '-v', '--version', action='version', version=VERSION
    )
    subparsers = parser.add_subparsers(title='modes')

    sub_parser = None
    for key, val in MODES.items():
        sub_parser = subparsers.add_parser(key, help=val['description'], add_help=False)

    if len(sys.argv) < 2:
        parser.print_help()
        exit()

    _ = parser.parse_known_args()
    sub_args = sub_parser.parse_known_args()

    mode = MODES[sub_args[1][0]]['main']
    sys.argv[0] = '{} {}'.format(SCRIPT_NAME, sys.argv[1])
    del sys.argv[1]
    exit(mode())



import pymongo


class Mongo:
    def __init__(self, conf):
        host = conf.d['mongo'].get('host', 'localhost')
        port = conf.d['mongo'].get('port', 27017)
        db = conf.d['mongo']['db']
        username = conf.d['mongo']['username']
        password = conf.d['mongo']['password']

        self.client = pymongo.MongoClient('mongodb://{username}:{password}@{host}:{port}/{db}'.format(
            username=username,
            password=password,
            host=host,
            port=port,
            db=db
        ))

        self.db = self.client[db]



import io
import json
import os
import tarfile
from threading import Thread, Event
import concurrent.futures
import time
from traceback import format_exc
from typing import List, Tuple, Dict

import docker
import jsonschema
from docker.models.containers import Container
from docker.tls import TLSConfig
from docker.types import Ulimit
import docker.errors
from bson.objectid import ObjectId
import bson.errors

from cc_agency.commons.schemas.callback import agent_result_schema
from cc_agency.commons.secrets import get_experiment_secret_keys, fill_experiment_secrets, fill_batch_secrets, \
    get_batch_secret_keys
from cc_core.commons.docker_utils import create_container_with_gpus
from cc_core.commons.engines import engine_to_runtime, NVIDIA_DOCKER_RUNTIME
from cc_core.commons.files import create_directory_tarinfo
from cc_core.commons.gpu_info import set_nvidia_environment_variables

from cc_agency.commons.helper import batch_failure
from cc_core.commons.red_to_blue import convert_red_to_blue, CONTAINER_OUTPUT_DIR, CONTAINER_AGENT_PATH, \
    CONTAINER_BLUE_FILE_PATH, CONTAINER_INPUT_DIR

INSPECTION_IMAGE = 'docker.io/busybox:latest'
NOFILE_LIMIT = 4096
CHECK_EXITED_CONTAINERS_INTERVAL = 1.0
OFFLINE_INSPECTION_INTERVAL = 10
CHECK_FOR_BATCHES_INTERVAL = 20


class ImagePullResult:
    def __init__(self, image_url, auth, successful, debug_info, depending_batches):
        """
        Creates a new DockerImagePull object.

        :param image_url: The url of the image to pull
        :type image_url: str
        :param auth: The authentication data for this image. Can be None if no authentication is required, otherwise it
                     has to be a tuple (username, password).
        :type auth: None or Tuple[str, str]
        :param successful: A boolean that is True, if the pull was successful
        :type successful: bool
        :param debug_info: A list of strings describing the error if the pull failed. Otherwise None.
        :type debug_info: List[str] or None
        :param depending_batches: A list of batches that depend on the execution of this docker pull
        :type depending_batches: List[Dict]
        """
        self.image_url = image_url
        self.auth = auth
        self.successful = successful
        self.debug_info = debug_info
        self.depending_batches = depending_batches


def _pull_image(docker_client, image_url, auth, depending_batches):
    """
    Pulls the given docker image and returns a ImagePullResult object.

    :param docker_client: The docker client, which is used to pull the image
    :type docker_client: docker.DockerClient
    :param image_url: The image to pull
    :type image_url: str
    :param auth: A tuple containing (username, password) or None
    :type auth: Tuple[str, str] or None
    :param depending_batches: A list of batches, which depend on the given image
    :type depending_batches: List[Dict]

    :return: An ImagePullResult describing the pull
    :rtype: ImagePullResult
    """
    try:
        docker_client.images.pull(image_url, auth_config=auth)
    except Exception as e:
        debug_info = str(e).split('\n')
        return ImagePullResult(image_url, auth, False, debug_info, depending_batches)

    return ImagePullResult(image_url, auth, True, None, depending_batches)


def _get_blue_agent_host_path():
    """
    :return: the absolute path to the blue agent of the host system
    :rtype: str
    """
    import cc_core.agent.blue.__main__ as blue_main
    return blue_main.__file__


class ClientProxy:
    """
    A client proxy handles a cluster node and the docker client of this node.
    It takes over the following tasks:
    - It queries the db for new batches, that got scheduled to this node to start them
    - It queries the docker client for containers, which are finished and updates the db accordingly
    - It queries the db to remove cancelled containers
    - If an error occurred it tries to reinitialize the docker client

    A client proxy contains a "online-flag" implemented as threading.Event. Any thread inside this client proxy except
    the inspection loop should wait for this event to be set, before starting a new execution cycle.
    Only the inspect-thread is allowed to set/clear the "online-flag" after successful/failed inspection.

    On error the check_for_batches and check_exited_containers-threads can trigger an inspection.

    inspect:
      If this ClientProxy is offline regularly inspects the connection to the docker daemon. If the inspection failed,
      the "online-flag" is cleared and this node is set to offline.

      If this ClientProxy is online inspect the docker engine, by starting a docker container and examine the result
      of this execution. If the execution was successful, set the "online-flag" and make this node online, otherwise
      repeat in some interval.

    check-for-batches:
      Regularly queries the database for batches, which are scheduled to this node. All found batches are then started
      with the docker client, if online. This can be triggered manually by setting the check-for-batches-event.
      If this client proxy is changed to be offline this thread processes the current cycle until it has finished and
      then waits for the "online-flag" to be set.

    check-exited-containers:
      Regularly queries the containers from the docker client and the database, which are currently running on this
      node. Checks the containers, which are not running anymore and handles their execution result. This can be
      triggered by setting the check_exited_containers-flag.
      If this client proxy is changed to be offline this thread processes the current cycle until it has finished and
      then waits for the "online-flag" to be set.
    """
    NUM_WORKERS = 4

    def __init__(self, node_name, conf, mongo, trustee_client, scheduling_event):
        self._node_name = node_name
        self._conf = conf
        self._mongo = mongo
        self._trustee_client = trustee_client

        self._scheduling_event = scheduling_event

        node_conf = conf.d['controller']['docker']['nodes'][node_name]
        self._base_url = node_conf['base_url']
        self._tls = False
        if 'tls' in node_conf:
            self._tls = TLSConfig(**node_conf['tls'])

        self._environment = node_conf.get('environment')
        self._network = node_conf.get('network')

        # create db entry for this node
        node = {
            'nodeName': node_name,
            'state': None,
            'history': [],
            'ram': None,
            'cpus': None
        }

        bson_node_id = self._mongo.db['nodes'].insert_one(node).inserted_id
        self._node_id = str(bson_node_id)

        # init docker client
        self._client = None
        self._runtimes = None
        self._online = Event()  # type: Event

        self._inspection_event = Event()  # type: Event
        self._check_for_batches_event = Event()  # type: Event
        self._check_exited_containers_event = Event()  # type: Event

        if not self._init_docker_client():
            self.do_inspect()
            self._set_offline(format_exc())

        Thread(target=self._inspection_loop).start()
        Thread(target=self._check_for_batches_loop).start()
        Thread(target=self._check_exited_containers_loop).start()

        # initialize Executor Pools
        self._pull_executor = concurrent.futures.ThreadPoolExecutor(max_workers=ClientProxy.NUM_WORKERS)
        self._run_executor = concurrent.futures.ThreadPoolExecutor(max_workers=ClientProxy.NUM_WORKERS)

    def is_online(self):
        return self._online.is_set()

    def _set_online(self, ram, cpus):
        print('Node online:', self._node_name)

        bson_node_id = ObjectId(self._node_id)
        self._mongo.db['nodes'].update_one(
            {'_id': bson_node_id},
            {
                '$set': {
                    'state': 'online',
                    'ram': ram,
                    'cpus': cpus
                },
                '$push': {
                    'history': {
                        'state': 'online',
                        'time': time.time(),
                        'debugInfo': None
                    }
                }
            }
        )

        self._online.set()  # start _check_batch_containers and _check_exited_containers

    def _set_offline(self, debug_info):
        print('Node offline:', self._node_name)

        self._online.clear()

        timestamp = time.time()
        bson_node_id = ObjectId(self._node_id)
        self._mongo.db['nodes'].update_one(
            {'_id': bson_node_id},
            {
                '$set': {'state': 'offline'},
                '$push': {
                    'history': {
                        'state': 'offline',
                        'time': timestamp,
                        'debugInfo': debug_info
                    }
                }
            }
        )

        # change state of assigned batches
        cursor = self._mongo.db['batches'].find(
            {
                'node': self._node_name,
                'state': {'$in': ['scheduled', 'processing']}
            },
            {'_id': 1, 'state': 1}
        )

        for batch in cursor:
            bson_id = batch['_id']
            batch_id = str(bson_id)
            debug_info = 'Node offline: {}'.format(self._node_name)
            batch_failure(self._mongo, batch_id, debug_info, None, batch['state'])

    def _info(self):
        info = self._client.info()
        ram = info['MemTotal'] // (1024 * 1024)
        cpus = info['NCPU']
        runtimes = info['Runtimes']
        return ram, cpus, runtimes

    def _batch_containers(self, status):
        """
        Returns a dictionary that maps container names to the corresponding container.
        If this client proxy is offline, the result will always be an empty dictionary.

        :param status: A status string. Containers, which have a different state are not contained in the result of this
                       function
        :type status: str or None
        :return: A dictionary mapping container names to docker containers
        :rtype: Dict[str, Container]

        :raise docker.errors.DockerException: If the docker engine returns an error
        """
        batch_containers = {}  # type: Dict[str, Container]

        if not self.is_online():
            return batch_containers

        filters = {'status': status}
        if status is None:
            filters = None

        containers = self._client.containers.list(all=True, limit=-1, filters=filters)  # type: List[Container]

        for c in containers:
            try:
                ObjectId(c.name)
                batch_containers[c.name] = c
            except (bson.errors.InvalidId, TypeError):
                pass

        return batch_containers

    def _fail_batches_without_assigned_container(self):
        containers = self._batch_containers(None)

        cursor = self._mongo.db['batches'].find(
            {
                'node': self._node_name,
                'state': {'$in': ['processing']}
            },
            {'_id': 1, 'state': 1}
        )

        for batch in cursor:
            bson_id = batch['_id']
            batch_id = str(bson_id)

            if batch_id not in containers:
                debug_info = 'No container assigned.'
                batch_failure(self._mongo, batch_id, debug_info, None, batch['state'])

    def _remove_cancelled_containers(self):
        """
        Stops all docker containers, whose batches got cancelled.

        :raise docker.errors.DockerException: If the docker server returns an error
        """
        running_containers = self._batch_containers('running')

        cursor = self._mongo.db['batches'].find(
            {
                '_id': {'$in': [ObjectId(_id) for _id in running_containers]},
                'state': 'cancelled'
            },
            {'_id': 1}
        )
        resources_freed = False
        for batch in cursor:
            batch_id = str(batch['_id'])

            c = running_containers[batch_id]
            c.remove(force=True)
            resources_freed = True

        return resources_freed

    def _can_execute_container(self):
        """
        Tries to execute a docker container using the docker client.

        :return: A tuple (successful, info/error)
                 successful: True, if the docker container could be executed, otherwise False
                 info/error: In case of success a tuple containing (ram, cpus, runtimes),
                 in case of failure the error message message as string.
        :rtype: tuple[bool, tuple or str]
        """
        command = ['echo', 'test']

        try:
            self._client.containers.run(
                INSPECTION_IMAGE,
                command,
                user='1000:1000',
                remove=True,
                environment=self._environment,
                network=self._network
            )
            info = self._info()
        except docker.errors.DockerException as e:
            return False, str(e)

        return True, info

    def _inspect_on_error(self):
        """
        Inspects the current docker client and sets this node to offline, if the inspection fails.
        """
        success, state = self._can_execute_container()

        if not success:
            self._set_offline(state)

    def _init_docker_client(self):
        """
        Tries to reinitialize the docker client. If successful, this node is online after this function execution.

        :return: True, if the initialization succeeded, otherwise False
        :rtype: bool
        """
        init_succeeded = False
        try:
            self._client = docker.DockerClient(base_url=self._base_url, tls=self._tls, version='auto')

            successful, state = self._can_execute_container()
            if successful:
                ram, cpus, runtimes = state
                self._runtimes = runtimes
                if not self.is_online():
                    self._set_online(ram, cpus)
                    init_succeeded = True
        except docker.errors.DockerException:
            pass
        return init_succeeded

    def _inspection_loop(self):
        """
        Regularly inspects the connection the docker daemon by running a docker container. If an error was found, clears
        the "online-flag" and puts in a error token in the inspection queue.

        Waits for errors inside the inspection-queue. If an error token was found inside the inspection-queue, handles
        this error. Otherwise performs a routine check of the docker client, after a given timeout.
        If this client proxy is offline, tries to restart it.
        """
        while True:
            if self.is_online():
                self._inspection_event.wait()
                self._inspection_event.clear()
                self._inspect_on_error()
            else:
                self._inspection_event.wait(timeout=OFFLINE_INSPECTION_INTERVAL)
                self._inspection_event.clear()
                self._init_docker_client()  # tries to reinitialize the docker client

    def _check_exited_containers(self):
        """
        Checks all containers which have "recently exited". A container is considered "recently exited", if the docker
        container is in state 'exited' and the corresponding batch is in state 'processing'.

        After this function execution all "recently exited" containers of this docker client should have batches, which
        are in one of the following states:
        - succeeded: If the batch execution was successful
        - failed: If the batch execution has failed

        :return: True if a exited container was found, otherwise False
        :rtype: bool

        :raise docker.errors.DockerException:
        """
        exited_containers = self._batch_containers('exited')  # type: Dict[str, Container]

        batch_cursor = self._mongo.db['batches'].find(
            {'_id': {'$in': [ObjectId(_id) for _id in exited_containers]}},
            {'state': 1}
        )
        resources_freed = False
        for batch in batch_cursor:
            batch_id = str(batch['_id'])

            exited_container = exited_containers[batch_id]

            self._check_exited_container(exited_container, batch)

            exited_container.remove()

            resources_freed = True

        return resources_freed

    def _check_exited_containers_loop(self):
        """
        Regularly checks exited containers. Waits for this client proxy to come online, before starting a new cycle.
        Also removes containers, whose batches got cancelled.
        """
        while True:
            self._online.wait()  # wait for this node to come online

            self._check_exited_containers_event.wait(timeout=CHECK_EXITED_CONTAINERS_INTERVAL)
            self._check_exited_containers_event.clear()

            try:
                resources_freed = self._check_exited_containers()
                resources_freed = self._remove_cancelled_containers() or resources_freed

                if resources_freed:
                    self._scheduling_event.set()
            except docker.errors.DockerException:
                self.do_inspect()

    def _check_exited_container(self, container, batch):
        """
        Inspects the logs of the given exited container and updates the database accordingly.

        :param container: The container to inspect
        :type container: Container
        :param batch: The batch to update according to the result of the container execution.
        :type batch: dict
        """
        bson_batch_id = batch['_id']
        batch_id = str(bson_batch_id)

        try:
            stdout_logs = container.logs(stderr=False).decode('utf-8')
            stderr_logs = container.logs(stdout=False).decode('utf-8')
            docker_stats = container.stats(stream=False)
        except Exception as e:
            debug_info = 'Could not get logs or stats of container: {}'.format(str(e))
            batch_failure(self._mongo, batch_id, debug_info, None, batch['state'])
            return

        data = None
        try:
            data = json.loads(stdout_logs)
        except json.JSONDecodeError as e:
            debug_info = 'CC-Agent data is not a valid json object: {}\n\nstdout was:\n{}'.format(str(e), stdout_logs)
            batch_failure(self._mongo, batch_id, debug_info, data, batch['state'], docker_stats=docker_stats)
            return

        try:
            jsonschema.validate(data, agent_result_schema)
        except jsonschema.ValidationError as e:
            debug_info = 'CC-Agent data sent by callback does not comply with jsonschema: {}'.format(str(e))
            batch_failure(self._mongo, batch_id, debug_info, data, batch['state'], docker_stats=docker_stats)
            return

        if data['state'] == 'failed':
            debug_info = 'Batch failed.\nContainer stderr:\n{}\ndebug info:\n{}'.format(stderr_logs, data['debugInfo'])
            batch_failure(self._mongo, batch_id, debug_info, data, batch['state'], docker_stats=docker_stats)
            return

        batch = self._mongo.db['batches'].find_one(
            {'_id': bson_batch_id, 'state': 'processing'},
            {'attempts': 1, 'node': 1}
        )
        if not batch:
            raise ValueError('batch id "{}" not found.'.format(batch_id))

        self._mongo.db['batches'].update_one(
            {
                '_id': bson_batch_id,
                'state': 'processing'
            },
            {
                '$set': {
                    'state': 'succeeded'
                },
                '$push': {
                    'history': {
                        'state': 'succeeded',
                        'time': time.time(),
                        'debugInfo': None,
                        'node': batch['node'],
                        'ccagent': data,
                        'dockerStats': docker_stats
                    }
                }
            }
        )

    def do_check_for_batches(self):
        """
        Triggers a check-for-batches cycle.
        """
        self._check_for_batches_event.set()

    def do_check_exited_containers(self):
        """
        Triggers a check-exited-containers cycle.
        """
        self._check_exited_containers_event.set()

    def do_inspect(self):
        """
        Triggers an inspection cycle.
        """
        self._inspection_event.set()

    def _check_for_batches_loop(self):
        """
        Regularly calls _check_for_batches. Does wait before executing a new cycle, if this client proxy is offline.
        """
        while True:
            self._online.wait()

            self._check_for_batches_event.wait(timeout=CHECK_FOR_BATCHES_INTERVAL)
            self._check_for_batches_event.clear()

            try:
                self._check_for_batches()
            except TrusteeServiceError:
                self.do_inspect()

    def _check_for_batches(self):
        """
        Queries the database to find batches, which are in state 'scheduled' and are scheduled to the node of this
        ClientProxy.
        First all docker images are pulled, which are used to process these batches. Afterwards the batch processing is
        run. The state in the database for these batches is then updated to 'processing'.

        :raise TrusteeServiceError: If the trustee service is unavailable or the trustee service could not fulfill all
        requested keys
        :raise ImageAuthenticationError: If the image authentication information is invalid.
        """

        # query for batches, that are in state 'scheduled' and are scheduled to this node
        query = {
            'state': 'scheduled',
            'node': self._node_name
        }

        # list containing batches that are scheduled to this node and save them together with their experiment
        batches_with_experiments = []  # type: List[Tuple[Dict, Dict]]

        # dictionary, that maps docker image authentications to batches, which need this docker image
        image_to_batches = {}  # type: Dict[Tuple, List[Dict]]

        for batch in self._mongo.db['batches'].find(query):
            experiment = self._get_experiment_with_secrets(batch['experimentId'])
            batches_with_experiments.append((batch, experiment))

            image_authentication = ClientProxy._get_image_authentication(experiment)
            if image_authentication not in image_to_batches:
                image_to_batches[image_authentication] = []
            image_to_batches[image_authentication].append(batch)

        # pull images
        pull_futures = []
        for image_authentication, depending_batches in image_to_batches.items():
            image_url, auth = image_authentication
            future = self._pull_executor.submit(_pull_image, self._client, image_url, auth, depending_batches)
            pull_futures.append(future)

        for pull_future in pull_futures:
            image_pull_result = pull_future.result()  # type: ImagePullResult

            # If pulling failed, the batches, which needed this image fail and are removed from the
            # batches_with_experiments list
            if not image_pull_result.successful:
                for batch in image_pull_result.depending_batches:
                    # fail the batch
                    batch_id = str(batch['_id'])
                    self._pull_image_failure(image_pull_result.debug_info, batch_id, batch['state'])

                    # remove batches that are failed
                    batches_with_experiments = list(filter(
                        lambda batch_with_experiment: str(batch_with_experiment[0]['_id']) != batch_id,
                        batches_with_experiments
                    ))

        # run every batch, that has not failed
        run_futures = []  # type: List[concurrent.futures.Future]
        for batch, experiment in batches_with_experiments:
            future = self._run_executor.submit(
                ClientProxy._run_batch_container_and_handle_exceptions,
                self,
                batch,
                experiment
            )
            run_futures.append(future)

        # wait for all batches to run
        concurrent.futures.wait(run_futures, return_when=concurrent.futures.ALL_COMPLETED)

    def _get_experiment_with_secrets(self, experiment_id):
        """
        Returns the experiment of the given experiment_id with filled secrets.

        :param experiment_id: The experiment id to resolve.
        :type experiment_id: ObjectId
        :return: The experiment as dictionary with filled template values.
        :raise TrusteeServiceError: If the trustee service is unavailable or the trustee service could not fulfill all
        requested keys
        """
        experiment = self._mongo.db['experiments'].find_one(
            {'_id': ObjectId(experiment_id)},
        )

        experiment = self._fill_experiment_secret_keys(experiment)

        return experiment

    def _fill_experiment_secret_keys(self, experiment):
        """
        Returns the given experiment with filled template keys and values.

        :param experiment: The experiment to complete.
        :return: Returns the given experiment with filled template keys and values.
        :raise TrusteeServiceError: If the trustee service is unavailable or the trustee service could not fulfill all
        requested keys
        """
        experiment_secret_keys = get_experiment_secret_keys(experiment)
        response = self._trustee_client.collect(experiment_secret_keys)
        if response['state'] == 'failed':

            debug_info = response['debugInfo']

            if response.get('inspect'):
                response = self._trustee_client.inspect()
                if response['state'] == 'failed':
                    debug_info = response['debug_info']
                    raise TrusteeServiceError('Trustee service unavailable:{}{}'.format(os.linesep, debug_info))

            experiment_id = str(experiment['_id'])
            raise TrusteeServiceError(
                'Trustee service request failed for experiment "{}":{}{}'.format(experiment_id, os.linesep, debug_info)
            )

        experiment_secrets = response['secrets']
        return fill_experiment_secrets(experiment, experiment_secrets)

    @staticmethod
    def _get_image_url(experiment):
        """
        Gets the url of the docker image for the given experiment

        :param experiment: The experiment whose docker image url is returned
        :type experiment: Dict
        :return: The url of the docker image for the given experiment
        """
        return experiment['container']['settings']['image']['url']

    @staticmethod
    def _get_image_authentication(experiment):
        """
        Reads the docker authentication information from the given experiment and returns it as tuple. The first element
        is always the image_url for the docker image. The second element is a tuple containing the username and password
        for authentication at the docker registry. If no username and password is given, the second return value is None

        :param experiment: An experiment with filled secret keys, whose image authentication information should be
        returned
        :type experiment: Dict

        :return: A tuple containing the image_url as first element. The second element can be None or a Tuple containing
        (username, password) for authentication at the docker registry.
        :rtype: Tuple[str, None] or Tuple[str, Tuple[str, str]]

        :raise ImageAuthenticationError: If the given image authentication information is not complete
        (username and password are mandatory)
        """

        image_url = ClientProxy._get_image_url(experiment)

        image_auth = experiment['container']['settings']['image'].get('auth')
        if image_auth:
            for mandatory_key in ('username', 'password'):
                if mandatory_key not in image_auth:
                    raise ImageAuthenticationError(
                        'Image authentication is given, but "{}" is missing'.format(mandatory_key)
                    )

            image_auth = (image_auth['username'], image_auth['password'])
        else:
            image_auth = None

        return image_url, image_auth

    def _run_batch_container_and_handle_exceptions(self, batch, experiment):
        """
        Runs the given batch by calling _run_batch_container(), but handles exceptions, by calling
        _run_batch_container_failure().

        :param batch: The batch to run
        :type batch: dict
        :param experiment: The experiment of this batch
        :type experiment: dict
        :return:
        """
        try:
            self._run_batch_container(batch, experiment)
        except Exception as e:
            batch_id = str(batch['_id'])
            self._run_batch_container_failure(batch_id, str(e), batch['state'])

    def _run_batch_container(self, batch, experiment):
        """
        Creates a docker container and runs the given batch, with settings described in the given batch and experiment.
        Sets the state of the given batch to 'processing'.

        :param batch: The batch to run
        :type batch: dict
        :param experiment: The experiment of this batch
        :type experiment: dict
        """
        batch_id = str(batch['_id'])

        update_result = self._mongo.db['batches'].update_one(
            {
                '_id': ObjectId(batch_id),
                'state': 'scheduled'
            },
            {
                '$set': {
                    'state': 'processing',
                },
                '$push': {
                    'history': {
                        'state': 'processing',
                        'time': time.time(),
                        'debugInfo': None,
                        'node': self._node_name,
                        'ccagent': None,
                        'dockerStats': None
                    }
                }
            }
        )

        # only run the docker container, if the batch was successfully updated
        if update_result.modified_count == 1:
            self._run_container(batch, experiment)

    def _run_container(self, batch, experiment):
        """
        Runs a docker container for the given batch. Uses the following procedure:

        - Collects all arguments for the docker container execution
        - Removes old containers with the same name
        - Creates the docker container with the collected arguments
        - Creates an archive containing the blue_agent and the blue_file of this batch and copies this archive into the
          container
        - Starts the container

        :param batch: The batch to run inside the container
        :type batch: Dict[str, Any]
        :param experiment: The experiment of the given batch
        :type experiment: Dict[str, Any]
        """
        batch_id = str(batch['_id'])

        environment = {}
        if self._environment:
            environment = self._environment.copy()

        gpus = batch['usedGPUs']

        # set mount variables
        devices = []
        capabilities = []
        security_opt = []
        if batch['mount']:
            devices.append('/dev/fuse')
            capabilities.append('SYS_ADMIN')
            security_opt.append('apparmor:unconfined')

        # set image
        image = experiment['container']['settings']['image']['url']

        command = [
            'python3',
            CONTAINER_AGENT_PATH,
            '--outputs',
            '--debug',
            CONTAINER_BLUE_FILE_PATH
        ]

        ram = experiment['container']['settings']['ram']
        mem_limit = '{}m'.format(ram)

        # set ulimits
        ulimits = [
            docker.types.Ulimit(
                name='nofile',
                soft=NOFILE_LIMIT,
                hard=NOFILE_LIMIT
            )
        ]

        # remove container if it exists from earlier attempt
        existing_container = self._batch_containers(None).get(batch_id)
        if existing_container is not None:
            existing_container.remove(force=True)

        container = create_container_with_gpus(
            client=self._client,
            image=image,
            command=command,
            available_runtimes=self._runtimes,
            name=batch_id,
            user='1000:1000',
            working_dir=CONTAINER_OUTPUT_DIR,
            detach=True,
            mem_limit=mem_limit,
            memswap_limit=mem_limit,
            gpus=gpus,
            environment=environment,
            network=self._network,
            devices=devices,
            cap_add=capabilities,
            security_opt=security_opt,
            ulimits=ulimits
        )  # type: Container

        # copy blue agent and blue file to container
        tar_archive = self._create_batch_archive(batch)
        container.put_archive('/', tar_archive)
        tar_archive.close()

        container.start()

    def _create_blue_batch(self, batch):
        """
        Creates a dictionary containing the data for a blue batch.

        :param batch: The batch description
        :type batch: dict
        :return: A dictionary containing a blue batch
        :rtype: dict
        :raise TrusteeServiceError: If the trustee service is unavailable or unable to collect the requested secret keys
        :raise ValueError: If there was more than one blue batch after red_to_blue
        """
        batch_id = str(batch['_id'])
        batch_secret_keys = get_batch_secret_keys(batch)
        response = self._trustee_client.collect(batch_secret_keys)

        if response['state'] == 'failed':
            debug_info = 'Trustee service failed:\n{}'.format(response['debug_info'])
            disable_retry = response.get('disable_retry')
            batch_failure(
                self._mongo,
                batch_id,
                debug_info,
                None,
                batch['state'],
                disable_retry_if_failed=disable_retry
            )
            raise TrusteeServiceError(debug_info)

        batch_secrets = response['secrets']
        batch = fill_batch_secrets(batch, batch_secrets)

        experiment_id = batch['experimentId']

        experiment = self._mongo.db['experiments'].find_one(
            {'_id': ObjectId(experiment_id)}
        )

        red_data = {
            'redVersion': experiment['redVersion'],
            'cli': experiment['cli'],
            'inputs': batch['inputs'],
            'outputs': batch['outputs']
        }

        blue_batches = convert_red_to_blue(red_data)

        if len(blue_batches) != 1:
            raise ValueError('Got {} batches, but only one was asserted.'.format(len(blue_batches)))

        return blue_batches[0]

    def _create_batch_archive(self, batch):
        """
        Creates a tar archive.
        This archive contains the blue agent, a blue file, the outputs-directory and the inputs-directory.
        The blue file is filled with the blue data from the given batch.
        The outputs-directory is an empty directory, with name 'outputs'
        The inputs-directory is an empty directory, with name 'inputs'
        The tar archive and the blue file are always in memory and never stored on the local filesystem.

        The resulting archive is:
        /cc
        |--/blue_agent.py
        |--/blue_file.json
        |--/outputs/
        |--/inputs/

        :param batch: The data to put into the blue file of the returned archive
        :type batch: dict
        :return: A tar archive containing the blue agent and the given blue batch
        :rtype: io.BytesIO or bytes
        """
        data_file = io.BytesIO()
        tar_file = tarfile.open(mode='w', fileobj=data_file)

        # add blue agent
        tar_file.add(_get_blue_agent_host_path(), arcname=CONTAINER_AGENT_PATH, recursive=False)

        # add blue file. See https://bugs.python.org/issue22208 for more information
        blue_batch = self._create_blue_batch(batch)
        blue_batch_content = json.dumps(blue_batch).encode('utf-8')

        blue_batch_tarinfo = tarfile.TarInfo(CONTAINER_BLUE_FILE_PATH)
        blue_batch_tarinfo.size = len(blue_batch_content)

        tar_file.addfile(blue_batch_tarinfo, io.BytesIO(blue_batch_content))

        # add outputs directory
        output_directory_tarinfo = create_directory_tarinfo(CONTAINER_OUTPUT_DIR, owner_name='cc')
        tar_file.addfile(output_directory_tarinfo)

        # add inputs directory
        input_directory_tarinfo = create_directory_tarinfo(CONTAINER_INPUT_DIR, owner_name='cc')
        tar_file.addfile(input_directory_tarinfo)

        # close file
        tar_file.close()
        data_file.seek(0)

        return data_file

    def _run_batch_container_failure(self, batch_id, debug_info, current_state):
        batch_failure(self._mongo, batch_id, debug_info, None, current_state)

    def _pull_image_failure(self, debug_info, batch_id, current_state):
        batch_failure(self._mongo, batch_id, debug_info, None, current_state)


class TrusteeServiceError(Exception):
    pass


class ImageAuthenticationError(Exception):
    pass



from time import time

from cc_core.commons.red import red_validation
from flask import jsonify, request
from werkzeug.exceptions import Unauthorized, BadRequest, NotFound, InternalServerError
from bson.objectid import ObjectId

from cc_core.commons.red import red_validation
from cc_core.commons.engines import engine_validation
from cc_core.commons.templates import get_template_keys, get_secret_values, normalize_keys
from cc_core.commons.exceptions import exception_format
from cc_core.commons.red_to_blue import convert_red_to_blue

from cc_agency.commons.helper import str_to_bool
from cc_agency.commons.secrets import separate_secrets_batch, separate_secrets_experiment


def _prepare_red_data(data, user):
    timestamp = time()

    experiment = {
        'username': user['username'],
        'registrationTime': timestamp,
        'redVersion': data['redVersion'],
        'cli': data['cli'],
        'container': data['container'],
        'protectedKeysVoided': False
    }

    if 'execution' in data:
        stripped_settings = {}

        for key, val in data['execution']['settings'].items():
            if key == 'access':
                continue

            stripped_settings[key] = val

        experiment['execution'] = {
            'engine': data['execution']['engine'],
            'settings': stripped_settings
        }

    experiment, secrets = separate_secrets_experiment(experiment)

    if 'batches' in data:
        raw_batches = data['batches']
    else:
        raw_batches = [{
            'inputs': data['inputs'],
            'outputs': data['outputs']
        }]

    batches = []

    for i, rb in enumerate(raw_batches):
        batch = {
            'username': user['username'],
            'registrationTime': timestamp,
            'state': 'registered',
            'batchesListIndex': i,
            'protectedKeysVoided': False,
            'notificationsSent': False,
            'node': None,
            'history': [{
                'state': 'registered',
                'time': timestamp,
                'debugInfo': None,
                'node': None,
                'ccagent': None,
                'dockerStats': None
            }],
            'attempts': 0,
            'inputs': rb['inputs'],
            'outputs': rb['outputs']
        }
        batch, additional_secrets = separate_secrets_batch(batch)
        secrets.update(additional_secrets)
        batches.append(batch)

    return experiment, batches, secrets


def red_routes(app, mongo, auth, controller, trustee_client):
    @app.route('/red', methods=['POST'])
    def post_red():
        user = auth.verify_user(request.authorization)
        if not user:
            raise Unauthorized()

        if not request.json:
            raise BadRequest('Did not send RED data as JSON.')

        data = request.json

        try:
            red_validation(data, False)
        except Exception as e:
            raise BadRequest(
                'Given RED data is invalid. Consider using the FAICE commandline tools for local validation.\n{}'
                .format(str(e))
            )

        template_keys = set()
        get_template_keys(data, template_keys)
        if template_keys:
            raise BadRequest(
                'The given red data contains the following variables: "{}". Please resolve them before submitting'
                ' to agency. Consider using CC-FAICE (faice exec).'.format(', '.join(template_keys)))

        secret_values = get_secret_values(data)

        if 'batches' in data:
            for batch in data['batches']:
                if 'outputs' not in batch:
                    raise BadRequest(
                        'CC-Agency requires all batches to have outputs defined. At least one batch does not comply.'
                    )

        elif 'outputs' not in data:
            raise BadRequest('CC-Agency requires outputs to be defined in RED data.')

        try:
            engine_validation(data, 'container', ['docker'])
        except Exception:
            raise BadRequest('\n'.join(exception_format(secret_values=secret_values)))

        if 'ram' not in data['container']['settings']:
            raise BadRequest('CC-Agency requires \'ram\' to be defined in the container settings.')

        try:
            engine_validation(data, 'execution', ['ccagency'], optional=True)
            normalize_keys(data)
            _ = convert_red_to_blue(data)
        except Exception:
            raise BadRequest('\n'.join(exception_format(secret_values=secret_values)))

        experiment, batches, secrets = _prepare_red_data(data, user)

        response = trustee_client.store(secrets)
        if response['state'] == 'failed':
            raise InternalServerError('Trustee service failed:\n{}'.format(response['debug_info']))

        bson_experiment_id = mongo.db['experiments'].insert_one(experiment).inserted_id
        experiment_id = str(bson_experiment_id)

        for batch in batches:
            batch['experimentId'] = experiment_id

        mongo.db['batches'].insert_many(batches)

        controller.send_json({'destination': 'scheduler'})

        return jsonify({'experimentId': experiment_id})

    @app.route('/batches/<object_id>', methods=['DELETE'])
    def delete_batches(object_id):
        user = auth.verify_user(request.authorization)
        if not user:
            raise Unauthorized()

        try:
            bson_id = ObjectId(object_id)
        except Exception:
            raise BadRequest('Not a valid BSON ObjectId.')

        match = {'_id': bson_id}
        match_with_state = {'_id': bson_id, 'state': {'$nin': ['succeeded', 'failed', 'cancelled']}}

        if not user['is_admin']:
            match['username'] = user['username']
            match_with_state['username'] = user['username']

        o = mongo.db['batches'].find_one(match, {'state': 1})
        if not o:
            raise NotFound('Could not find Object.')

        mongo.db['batches'].update_one(
            match_with_state,
            {
                '$set': {
                    'state': 'cancelled'
                },
                '$push': {
                    'history': {
                        'state': 'cancelled',
                        'time': time(),
                        'debugInfo': None,
                        'node': None,
                        'ccagent': None,
                        'dockerStats': None
                    }
                }
            })

        o = mongo.db['batches'].find_one(match)
        o['_id'] = str(o['_id'])

        controller.send_json({'destination': 'scheduler'})

        return jsonify(o)

    @app.route('/experiments/count', methods=['GET'])
    def get_experiments_count():
        return get_collection_count('experiments')

    @app.route('/experiments', methods=['GET'])
    def get_experiments():
        return get_collection('experiments')

    @app.route('/experiments/<object_id>', methods=['GET'])
    def get_experiments_id(object_id):
        return get_collection_id('experiments', object_id)

    @app.route('/batches/count', methods=['GET'])
    def get_batches_count():
        return get_collection_count('batches')

    @app.route('/batches', methods=['GET'])
    def get_batches():
        return get_collection('batches')

    @app.route('/batches/<object_id>', methods=['GET'])
    def get_batches_id(object_id):
        return get_collection_id('batches', object_id)

    def get_collection_id(collection, object_id):
        user = auth.verify_user(request.authorization)
        if not user:
            raise Unauthorized()

        try:
            bson_id = ObjectId(object_id)
        except Exception:
            raise BadRequest('Not a valid BSON ObjectId.')

        match = {'_id': bson_id}

        if not user['is_admin']:
            match['username'] = user['username']

        o = mongo.db[collection].find_one(match)
        if not o:
            raise NotFound('Could not find Object.')

        o['_id'] = str(o['_id'])
        return jsonify(o)

    def get_collection_count(collection):
        user = auth.verify_user(request.authorization)
        if not user:
            raise Unauthorized()

        username = request.args.get('username', default=None, type=str)
        node = None
        experiment_id = None
        state = None

        if collection == 'batches':
            node = request.args.get('node', default=None, type=str)
            state = request.args.get('state', default=None, type=str)
            experiment_id = request.args.get('experimentId', default=None, type=str)

            if experiment_id:
                try:
                    _ = ObjectId(experiment_id)
                except Exception:
                    raise BadRequest('Experiment is not a valid BSON ObjectId.')

        aggregate = []

        if not user['is_admin']:
            aggregate.append({'$match': {'username': user['username']}})

        match = {}

        if username:
            match['username'] = username

        if node:
            match['node'] = node

        if experiment_id:
            match['experimentId'] = experiment_id

        if state:
            match['state'] = state

        aggregate.append({'$match': match})
        aggregate.append({'$count': 'count'})

        cursor = mongo.db[collection].aggregate(aggregate)
        cursor = list(cursor)
        
        if not cursor:
            return jsonify({'count': 0})

        return jsonify(cursor[0])

    def get_collection(collection):
        user = auth.verify_user(request.authorization)
        if not user:
            raise Unauthorized()

        skip = request.args.get('skip', default=None, type=int)
        limit = request.args.get('limit', default=None, type=int)
        username = request.args.get('username', default=None, type=str)
        ascending = str_to_bool(request.args.get('ascending', default=None, type=str))

        node = None
        experiment_id = None
        state = None

        if collection == 'batches':
            node = request.args.get('node', default=None, type=str)
            state = request.args.get('state', default=None, type=str)

            states = ['registered', 'scheduled', 'processing', 'succeeded', 'failed', 'cancelled']
            if state and state not in states:
                raise BadRequest('Given state is not valid. Must be one of {}.'.format(states))

            experiment_id = request.args.get('experimentId', default=None, type=str)

            if experiment_id:
                try:
                    _ = ObjectId(experiment_id)
                except Exception:
                    raise BadRequest('Experiment is not a valid BSON ObjectId.')

        aggregate = []

        if not user['is_admin']:
            aggregate.append({'$match': {'username': user['username']}})

        match = {}

        if username:
            match['username'] = username

        if node:
            match['node'] = node

        if experiment_id:
            match['experimentId'] = experiment_id

        if state:
            match['state'] = state

        aggregate.append({'$match': match})

        aggregate.append({'$project': {
            'username': 1,
            'registrationTime': 1,
            'state': 1,
            'experimentId': 1,
            'node': 1,
            'batchesListIndex': 1
        }})

        aggregate.append({'$sort': {'registrationTime': 1 if ascending else -1}})

        if skip is not None:
            if skip < 0:
                raise BadRequest('skip cannot be lower than 0.')
            aggregate.append({'$skip': skip})

        if limit is not None:
            if limit < 1:
                raise BadRequest('limit cannot be lower than 1.')
            aggregate.append({'$limit': limit})

        cursor = mongo.db[collection].aggregate(aggregate)

        result = []
        for e in cursor:
            e['_id'] = str(e['_id'])
            result.append(e)

        return jsonify(result)



from os import urandom
from binascii import hexlify
from time import time

from flask import request
from bson.objectid import ObjectId
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


def get_ip():
    headers = ['HTTP_X_FORWARDED_FOR', 'HTTP_X_REAL_IP', 'REMOTE_ADDR']
    ip = None
    for header in headers:
        ip = request.environ.get(header)
        if ip:
            break
    if not ip:
        ip = '127.0.0.1'
    return ip


def generate_secret():
    return hexlify(urandom(24)).decode('utf-8')


def create_kdf(salt):
    return PBKDF2HMAC(
        algorithm=SHA256(),
        length=32,
        salt=salt,
        iterations=100000,
        backend=default_backend()
    )


def batch_failure(
        mongo,
        batch_id,
        debug_info,
        ccagent,
        current_state,
        disable_retry_if_failed=False,
        docker_stats=None
):
    """
    Changes the db entry of the given batch to failed, if disable_retry_if_failed is set to True or if the maximal
    number of retries is exceeded. Otherwise the new state of the given batch is set to registered.

    :param mongo: The mongodb client to update
    :param batch_id: The batch id specifying the batch to fail
    :type batch_id: str
    :param debug_info: The debug info to write to the db
    :param ccagent: The ccagent to write to the db
    :param current_state: The expected current state of the batch to cancel. If this state does not match the batch from
                          the db, the db entry is not updated.
    :type current_state: str
    :param disable_retry_if_failed: If set to True, the batch is failed immediately, without giving another attempt
    :param docker_stats: The optional stats of the docker container, that will written under the "docker_stats" key in
                         the history of this batch
    :type docker_stats: dict
    """
    bson_id = ObjectId(batch_id)

    batch = mongo.db['batches'].find_one(
        {'_id': bson_id},
        {'attempts': 1, 'node': 1, 'experimentId': 1}
    )

    timestamp = time()
    attempts = batch['attempts']
    node_name = batch['node']

    new_state = 'registered'
    new_node = None

    if attempts >= 2 or disable_retry_if_failed:
        new_state = 'failed'
        new_node = node_name
    else:
        experiment_id = batch['experimentId']
        bson_experiment_id = ObjectId(experiment_id)
        experiment = mongo.db['experiments'].find_one(
            {'_id': bson_experiment_id},
            {'execution.settings.retryIfFailed': 1}
        )
        if not (experiment and experiment.get('execution', {}).get('settings', {}).get('retryIfFailed')):
            new_state = 'failed'
            new_node = node_name

    mongo.db['batches'].update_one(
        {'_id': bson_id, 'state': current_state},
        {
            '$set': {
                'state': new_state,
                'node': new_node
            },
            '$push': {
                'history': {
                    'state': new_state,
                    'time': timestamp,
                    'debugInfo': debug_info,
                    'node': new_node,
                    'ccagent': ccagent,
                    'dockerStats': docker_stats
                }
            }
        }
    )


def str_to_bool(s):
    if isinstance(s, str) and s.lower() in ['1', 'true']:
        return True
    return False



from flask import jsonify, request
from werkzeug.exceptions import Unauthorized
from bson.objectid import ObjectId


def _get_node_gpu_info(conf_nodes, node_name):
    node = conf_nodes.get(node_name)
    if node:
        hardware = node.get('hardware')
        if hardware:
            gpus = hardware.get('gpus')
            if gpus:
                return gpus
    return None


def nodes_routes(app, mongo, auth, conf):
    @app.route('/nodes', methods=['GET'])
    def get_nodes():
        conf_nodes = conf.d['controller']['docker']['nodes']
        user = auth.verify_user(request.authorization)
        if not user:
            raise Unauthorized()

        cursor = mongo.db['nodes'].find()

        nodes = list(cursor)
        node_names = [node['nodeName'] for node in nodes]

        cursor = mongo.db['batches'].find(
            {
                'node': {'$in': node_names},
                'state': {'$in': ['scheduled', 'processing']}
            },
            {'experimentId': 1, 'node': 1}
        )
        batches = list(cursor)
        experiment_ids = list(set([ObjectId(b['experimentId']) for b in batches]))

        cursor = mongo.db['experiments'].find(
            {'_id': {'$in': experiment_ids}},
            {'container.settings.ram': 1}
        )
        experiments = {str(e['_id']): e for e in cursor}

        for node in nodes:
            batches_ram = [
                {
                    'batchId': str(b['_id']),
                    'ram': experiments[b['experimentId']]['container']['settings']['ram']
                }
                for b in batches
                if b['node'] == node['nodeName']
            ]
            node['currentBatches'] = batches_ram
            gpu_info = _get_node_gpu_info(conf_nodes, node['nodeName'])
            if gpu_info:
                node['gpus'] = gpu_info
            del node['_id']

        return jsonify(nodes)



from hmac import compare_digest
from argparse import ArgumentParser

from flask import Flask, jsonify, request
from werkzeug.exceptions import Unauthorized

from cc_agency.commons.conf import Conf

DESCRIPTION = 'CC-Agency Trustee.'

app = Flask('trustee')
application = app

parser = ArgumentParser(description=DESCRIPTION)
parser.add_argument(
    '-c', '--conf-file', action='store', type=str, metavar='CONF_FILE',
    help='CONF_FILE (yaml) as local path.'
)
args = parser.parse_args()


conf = Conf(args.conf_file)
username = conf.d['trustee']['username']
password = conf.d['trustee']['password']


secrets = {}


def _verify_user(auth):
    if not auth:
        raise Unauthorized()

    if not compare_digest(auth.username, username):
        raise Unauthorized()

    if not compare_digest(auth.password, password):
        raise Unauthorized()


@app.route('/', methods=['GET'])
def get_root():
    _verify_user(request.authorization)

    return jsonify({
        'state': 'success'
    })


@app.route('/secrets', methods=['POST'])
def post_secrets():
    _verify_user(request.authorization)

    data = request.json

    existing_keys = []

    for key in data:
        if key in secrets:
            existing_keys.append(key)

    if existing_keys:
        return jsonify({
            'state': 'failed',
            'debug_info': 'Keys already exist: {}'.format(existing_keys),
            'disable_retry': False,
            'inspect': False
        })

    secrets.update(data)

    return jsonify({
        'state': 'success'
    })


@app.route('/secrets', methods=['GET'])
def get_secrets():
    _verify_user(request.authorization)

    data = request.json

    collected = {}
    missing_keys = []

    for key in data:
        try:
            collected[key] = secrets[key]
        except KeyError:
            missing_keys.append(key)

    if missing_keys:
        return jsonify({
            'state': 'failed',
            'debug_info': 'Could not collect keys: {}'.format(missing_keys),
            'disable_retry': True,
            'inspect': False
        })

    return jsonify({
        'state': 'success',
        'secrets': collected
    })


@app.route('/secrets', methods=['DELETE'])
def delete_secrets():
    _verify_user(request.authorization)

    data = request.json

    for key in data:
        try:
            del secrets[key]
        except KeyError:
            pass

    return jsonify({
        'state': 'success'
    })



import os
import sys
from threading import Thread, Event
from time import time, sleep
from typing import Dict

import requests
from bson.objectid import ObjectId

from cc_core.commons.gpu_info import GPUDevice, match_gpus, get_gpu_requirements, InsufficientGPUError, \
    NVIDIA_GPU_VENDOR
from cc_core.commons.red import red_get_mount_connectors_from_inputs

from cc_agency.controller.docker import ClientProxy, TrusteeServiceError
from cc_agency.commons.helper import batch_failure
from cc_agency.commons.secrets import get_experiment_secret_keys, fill_experiment_secrets
from cc_agency.commons.secrets import get_batch_secret_keys

_CRON_INTERVAL = 60


class CompleteNode:
    """
    Represents a processing node inside a cluster.
    """

    def __init__(self, node_name, online, ram, gpus, ram_available, gpus_available, num_batches_running):
        """
        Initialises a new CompleteNode.
        :param node_name: The name of the node given by the agency config.
        :param online: Whether the given node is online or not.
        :param ram: The amount of ram of this node.
        :param gpus: The GPUs that are present on this node. Does include gpus, which are used by batches.
        :param ram_available: The ram that is available. Given by the amount of ram of the node minus the amount of ram
        used by batches.
        :param gpus_available: The GPUs that are available and not used by batches.
        :param num_batches_running: The number of batches currently running on the node.
        """
        self.node_name = node_name
        self.online = online
        self.ram = ram
        self.gpus = gpus
        self.ram_available = ram_available
        self.gpus_available = gpus_available
        self.num_batches_running = num_batches_running


class Scheduler:
    def __init__(self, conf, mongo, trustee_client):
        self._conf = conf
        self._mongo = mongo
        self._trustee_client = trustee_client

        mongo.db['nodes'].drop()

        self._scheduling_event = Event()
        self._voiding_event = Event()
        self._notification_event = Event()

        self._nodes = {
            node_name: ClientProxy(node_name, conf, mongo, trustee_client, self._scheduling_event)
            for node_name
            in conf.d['controller']['docker']['nodes'].keys()
        }  # type: Dict[str, ClientProxy]

        Thread(target=self._scheduling_loop).start()
        Thread(target=self._voiding_loop).start()
        Thread(target=self._notification_loop).start()

    def schedule(self):
        self._scheduling_event.set()

    def _notification_loop(self):
        while True:
            self._notification_event.wait()
            self._notification_event.clear()

            # batches
            cursor = self._mongo.db['batches'].find(
                {
                    'state': {'$in': ['succeeded', 'failed', 'cancelled']},
                    'notificationsSent': False
                },
                {'state': 1}
            )

            bson_ids = []
            payload = {'batches': []}

            for batch in cursor:
                bson_id = batch['_id']
                bson_ids.append(bson_id)

                payload['batches'].append({
                    'batchId': str(bson_id),
                    'state': batch['state']
                })

            self._mongo.db['batches'].update(
                {'_id': {'$in': bson_ids}},
                {'$set': {'notificationsSent': True}}
            )

            notification_hooks = self._conf.d['controller'].get('notification_hooks', [])

            for hook in notification_hooks:
                auth = hook.get('auth')

                if auth is not None:
                    auth = (auth['username'], auth['password'])

                try:
                    r = requests.post(hook['url'], auth=auth, json=payload)
                    r.raise_for_status()
                except Exception as e:
                    debug_info = 'Notification post hook failed:{0}{1}{0}{2}'.format(os.linesep, repr(e), e)
                    print(debug_info, file=sys.stderr)

    def _voiding_loop(self):
        while True:
            self._voiding_event.wait()
            self._voiding_event.clear()

            # batches
            cursor = self._mongo.db['batches'].find(
                {
                    'state': {'$in': ['succeeded', 'failed', 'cancelled']},
                    'protectedKeysVoided': False
                }
            )

            for batch in cursor:
                bson_id = batch['_id']

                batch_secret_keys = get_batch_secret_keys(batch)
                self._trustee_client.delete(batch_secret_keys)

                self._mongo.db['batches'].update_one({'_id': bson_id}, {'$set': {'protectedKeysVoided': True}})

            # experiments
            cursor = self._mongo.db['experiments'].find(
                {
                    'protectedKeysVoided': False
                }
            )

            for experiment in cursor:
                bson_id = experiment['_id']
                experiment_id = str(bson_id)

                all_count = self._mongo.db['batches'].count({'experimentId': experiment_id})

                finished_count = self._mongo.db['batches'].count({
                    'experimentId': experiment_id,
                    'state': {'$in': ['succeeded', 'failed', 'cancelled']}
                })

                if all_count == finished_count:
                    experiment_secret_keys = get_experiment_secret_keys(experiment)
                    self._trustee_client.delete(experiment_secret_keys)

                    self._mongo.db['experiments'].update_one({'_id': bson_id}, {'$set': {'protectedKeysVoided': True}})

    def _scheduling_loop(self):
        while True:
            self._scheduling_event.wait(timeout=_CRON_INTERVAL)
            self._scheduling_event.clear()

            # void protected keys
            self._voiding_event.set()

            # send notifications
            self._notification_event.set()

            # inspect trustee
            response = self._trustee_client.inspect()
            if response['state'] == 'failed':
                debug_info = response['debug_info']
                print('Trustee service unavailable, retry in {} seconds:{}{}'.format(
                    _CRON_INTERVAL, os.linesep, debug_info
                ), file=sys.stderr)
                sleep(_CRON_INTERVAL)
                continue

            self._client_proxies_check_exited_containers()
            self._schedule_batches()
            self._client_proxies_check_for_batches()

    def _client_proxies_check_exited_containers(self):
        """
        Triggers every client proxy to check for exited containers and cancelled batches.
        """
        for client_proxy in self._nodes.values():
            client_proxy.do_check_exited_containers()

    def _client_proxies_check_for_batches(self):
        """
        Triggers every client proxy to check for new batches possibly scheduled to their nodes.
        """
        for client_proxy in self._nodes.values():
            client_proxy.do_check_for_batches()

    @staticmethod
    def _get_busy_gpu_ids(batches, node_name):
        """
        Returns a list of busy GPUs in the given batches

        :param batches: The batches to analyse given as list of dictionaries.
                        If GPUs are busy by a current batch the key 'usedGPUs' should be present.
                        The value of 'usedGPUs' has to be a list of busy device IDs.
        :return: A list of GPUDevice-IDs, which are used by the given batches on the given node
        """

        busy_gpus = []
        for b in batches:
            if b['node'] == node_name:
                batch_gpus = b.get('usedGPUs')
                if type(batch_gpus) == list:
                    busy_gpus.extend(batch_gpus)

        return busy_gpus

    def _get_present_gpus(self, node_name):
        """
        Returns a list of GPUDevices

        :param node_name: The name of the node
        :return: A list of GPUDevices, which are representing the GPU Devices present on the specified node
        """

        result = []

        gpus = self._conf.d['controller']['docker']['nodes'][node_name].get('hardware', {}).get('gpus')
        if gpus:
            for gpu in gpus:
                result.append(GPUDevice(device_id=gpu['id'], vram=gpu['vram'], vendor=NVIDIA_GPU_VENDOR))

        return result

    def _get_available_gpus(self, node, batches):
        """
        Returns a list of available GPUs on the given node.
        Available in this context means, that this device is present on the node and is not busy with another batch.

        :param node: The node whose available GPUs should be calculated
        :param batches: The batches currently running
        :return: A list of available GPUDevices of the specified node
        """

        node_name = node['nodeName']

        busy_gpu_ids = Scheduler._get_busy_gpu_ids(batches, node_name)
        present_gpus = self._get_present_gpus(node_name)

        return [gpu for gpu in present_gpus if gpu.device_id not in busy_gpu_ids]

    def _get_cluster_state(self):
        """
        Returns a list of complete nodes, which are currently present in the cluster.
        :rtype: List[CompleteNode]
        """
        cursor = self._mongo.db['nodes'].find(
            {},
            {'state': 1, 'ram': 1, 'nodeName': 1}
        )

        nodes = list(cursor)
        node_names = [node['nodeName'] for node in nodes]

        cursor = self._mongo.db['batches'].find(
            {
                'node': {'$in': node_names},
                'state': {'$in': ['scheduled', 'processing']}},
            {'experimentId': 1, 'node': 1, 'usedGPUs': 1}
        )
        batches = list(cursor)
        experiment_ids = list(set([ObjectId(b['experimentId']) for b in batches]))

        cursor = self._mongo.db['experiments'].find(
            {'_id': {'$in': experiment_ids}},
            {'container.settings.ram': 1}
        )
        experiments = {str(e['_id']): e for e in cursor}

        complete_nodes = []

        for node in nodes:
            node_name = node['nodeName']
            node_batches = list(filter(lambda batch: batch['node'] == node_name, batches))

            num_batches = len(node_batches)

            used_ram = sum([
                experiments[b['experimentId']]['container']['settings']['ram']
                for b in node_batches
            ])

            available_gpus = self._get_available_gpus(node, batches)

            online = node['state'] == 'online'
            
            ram_available = None
            if node['ram'] is not None:
                ram_available = node['ram'] - used_ram

            complete_node = CompleteNode(
                node_name=node_name,
                online=online,
                ram=node['ram'],
                gpus=self._get_present_gpus(node['nodeName']),
                ram_available=ram_available,
                gpus_available=available_gpus,
                num_batches_running=num_batches,
            )

            complete_nodes.append(complete_node)

        return complete_nodes

    @staticmethod
    def _node_sufficient(node, experiment):
        """
        Returns True if the nodes hardware is sufficient for the experiment

        :param node: The node to test
        :type node: CompleteNode
        :param experiment: A dictionary containing hardware requirements for the experiment
        :return: True, if the nodes hardware is sufficient for the experiment, otherwise False
        """

        if not node.online:
            return False

        if node.ram_available < experiment['container']['settings']['ram']:
            return False

        # check gpus
        gpu_requirements = get_gpu_requirements(experiment['container']['settings'].get('gpus'))

        try:
            _gpus = match_gpus(node.gpus_available, gpu_requirements)
        except InsufficientGPUError:
            return False

        return True

    @staticmethod
    def _node_possibly_sufficient(node, experiment):
        """
        Returns True if the node could be sufficient for the experiment, even if the node does not have
        sufficient hardware at the moment (because of running batches).

        :param node: The node to check
        :type node: CompleteNode
        :param experiment: The experiment for which the node is sufficient or not.
        :return: True, if the node is possibly sufficient otherwise False
        """
        # check if node is initialized
        if (node.ram is None) or (node.gpus is None):
            return False

        if node.ram < experiment['container']['settings']['ram']:
            return False

        gpu_requirements = get_gpu_requirements(experiment['container']['settings'].get('gpus'))

        try:
            match_gpus(node.gpus, gpu_requirements)
        except InsufficientGPUError:
            return False
        return True

    @staticmethod
    def _check_nodes_possibly_sufficient(nodes, experiment):
        """
        Returns True if a possibly sufficient node is found otherwise False
        :param nodes: The nodes to check
        :type nodes: List[CompleteNode]
        :param experiment: The description of the experiment
        :return: True if a possibly sufficient node is found otherwise False
        """
        for node in nodes:
            if Scheduler._node_possibly_sufficient(node, experiment):
                return True
        return False

    @staticmethod
    def _get_best_node(nodes, experiment):
        """
        Returns the node, that fits best for the given experiment. If no node could be found returns None
        :param nodes: The nodes, that are available for this experiment.
        :type nodes: List[CompleteNode]
        :param experiment: The description of the experiment
        :return: The node that fits best for the given experiment. If no node fits at the moment None is returned.
        :rtype: CompleteNode
        """
        # check sufficient nodes
        sufficient_nodes = [node for node in nodes if Scheduler._node_sufficient(node, experiment)]
        if not sufficient_nodes:
            return None

        # prefer nodes without GPUs
        nodes_without_gpus = [node for node in sufficient_nodes if (not node.gpus)]
        if nodes_without_gpus:
            sufficient_nodes = nodes_without_gpus

        # prefer nodes with few jobs
        min_num_batches = None
        nodes_with_few_batches = []
        for node in sufficient_nodes:
            if min_num_batches is None:
                min_num_batches = node.num_batches_running
                nodes_with_few_batches.append(node)
                continue

            if node.num_batches_running < min_num_batches:
                min_num_batches = node.num_batches_running
                nodes_with_few_batches = [node]
            elif node.num_batches_running == min_num_batches:
                nodes_with_few_batches.append(node)

        # prefer nodes with less free ram
        nodes_with_few_batches.sort(reverse=False, key=lambda n: n.ram_available)

        return nodes_with_few_batches[0]

    def _schedule_batches(self):
        """
        state before _schedule_batches:
        There might be batches with state "registered" (given in _fifo()).
        There might be nodes, that are online and capable of processing the given batches (given in _online_nodes()).

        state after _schedule_batches:
        ClientProxies for which a batch is scheduled have a 'check_for_batches' action in their queue.
        Batches that are scheduled have state "scheduled" now and the node property of these batches is filled.
        """
        # list of tuple(batch_id, node_name) with node_names to which the batches were scheduled
        scheduled_nodes = []
        cluster_nodes = self._get_cluster_state()

        batch_count_cache = {}  # type: Dict[str, int]

        # select batch to be scheduled
        for next_batch in self._fifo():
            node_name = self._schedule_batch(next_batch, cluster_nodes, batch_count_cache)

            if node_name is not None:
                cluster_nodes = self._get_cluster_state()
                scheduled_nodes.append((next_batch['_id'], node_name))

        # inform ClientProxies about new batches
        for batch_id, node_name in scheduled_nodes:
            client_proxy = self._nodes[node_name]

            client_proxy.do_check_for_batches()

    def _get_number_of_batches_of_experiment(self, experiment_id, batch_count_cache):
        """
        Returns the number of batches, that are scheduled or processing of the given experiment.
        This number can be a overestimation of the real number.

        If the given experiment id is a key in batch_count_cache, the corresponding value is returned.
        If not the db is queried and an entry in batch_count_cache is created containing the queried data.

        After this function batch_count_cache always contains the experiment id as key.

        :param experiment_id: Defines the experiment, whose number of batches is returned
        :type experiment_id: str
        :param batch_count_cache: A dictionary mapping experiment ids to the number of batches of this experiment, which
                                  in state processing or scheduled. This dictionary is allowed to overestimate the
                                  number of batches.
        :type batch_count_cache: Dict[str, int]
        :return: The number of batches in state scheduled or processing of the given experiment id
        :rtype: int
        """
        if experiment_id in batch_count_cache:
            batch_count = batch_count_cache[experiment_id]
        else:
            batch_count = self._mongo.db['batches'].count({
                'experimentId': experiment_id,
                'state': {'$in': ['scheduled', 'processing']}
            })
            batch_count_cache[experiment_id] = batch_count
        return batch_count

    def _schedule_batch(self, next_batch, nodes, batch_count_cache):
        """
        Tries to find a node that is capable of processing the given batch. If no capable node could be found, None is
        returned.
        If a node was found, that is capable of processing the given batch, this node is written to the node property of
        the batch. The batches state is then updated to 'scheduled'.

        :param next_batch: The batch to schedule.
        :param nodes: The nodes on which the batch should be scheduled.
        :type nodes: List[CompleteNode]
        :param batch_count_cache: A dictionary mapping experiment ids to the number of batches of this experiment, which
                                  in state processing or scheduled. This dictionary is allowed to overestimate the
                                  number of batches.
        :type batch_count_cache: Dict[str, int]
        :return: The name of the node on which the given batch is scheduled
        If the batch could not be scheduled None is returned
        :raise TrusteeServiceError: If the trustee service is unavailable.
        """
        batch_id = str(next_batch['_id'])
        experiment_id = next_batch['experimentId']

        try:
            experiment = self._get_experiment_of_batch(experiment_id)
        except Exception as e:
            batch_failure(
                self._mongo,
                batch_id,
                repr(e),
                None,
                next_batch['state'],
                disable_retry_if_failed=True
            )
            return None

        ram = experiment['container']['settings']['ram']

        # limit the number of currently executed batches from a single experiment
        concurrency_limit = experiment.get('execution', {}).get('settings', {}).get('batchConcurrencyLimit', 64)

        # number of batches which are scheduled or processing of the given experiment
        batch_count = self._get_number_of_batches_of_experiment(experiment_id, batch_count_cache)

        if batch_count >= concurrency_limit:
            return None

        # check impossible experiments
        if not Scheduler._check_nodes_possibly_sufficient(nodes, experiment):
            debug_info = 'There are no nodes configured that are possibly sufficient for experiment "{}"' \
                .format(next_batch['experimentId'])
            batch_failure(
                self._mongo,
                batch_id,
                debug_info,
                None,
                next_batch['state'],
                disable_retry_if_failed=True
            )
            return None

        # select node
        selected_node = Scheduler._get_best_node(nodes, experiment)

        if selected_node is None:
            return None

        # calculate ram / gpus
        selected_node.ram_available -= ram

        used_gpu_ids = None
        if selected_node.gpus_available:
            gpu_requirements = get_gpu_requirements(experiment['container']['settings'].get('gpus'))
            available_gpus = selected_node.gpus_available
            used_gpus = match_gpus(available_gpus, requirements=gpu_requirements)

            used_gpu_ids = []
            for gpu in used_gpus:
                used_gpu_ids.append(gpu.device_id)
                available_gpus.remove(gpu)

        # check mounting
        mount_connectors = red_get_mount_connectors_from_inputs(next_batch['inputs'])
        is_mounting = bool(mount_connectors)

        allow_insecure_capabilities = self._conf.d['controller']['docker'].get('allow_insecure_capabilities', False)

        if not allow_insecure_capabilities and is_mounting:
            # set state to failed, because insecure_capabilities are not allowed but needed, by this batch.
            debug_info = 'FUSE support for this agency is disabled, but the following input/output-keys are ' \
                         'configured to mount inside a docker container.{}{}'.format(os.linesep, mount_connectors)
            batch_failure(
                self._mongo,
                batch_id,
                debug_info,
                None,
                next_batch['state'],
                disable_retry_if_failed=True
            )
            return None

        # update batch data
        update_result = self._mongo.db['batches'].update_one(
            {'_id': next_batch['_id'], 'state': next_batch['state']},
            {
                '$set': {
                    'state': 'scheduled',
                    'node': selected_node.node_name,
                    'usedGPUs': used_gpu_ids,
                    'mount': is_mounting
                },
                '$push': {
                    'history': {
                        'state': 'scheduled',
                        'time': time(),
                        'debugInfo': None,
                        'node': selected_node.node_name,
                        'ccagent': None,
                        'dockerStats': None
                    }
                },
                '$inc': {
                    'attempts': 1
                }
            }
        )

        if update_result.modified_count == 1:
            # The state of the scheduled batch switched from 'registered' to 'scheduled', so increase the batch_count.
            # batch_count_cache always contains experiment_id, because _get_number_of_batches_of_experiment()
            # always inserts the given experiment_id
            batch_count_cache[experiment_id] += 1

            return selected_node.node_name
        else:
            return None

    def _get_experiment_of_batch(self, experiment_id):
        """
        Returns the experiment of the given experiment_id with filled secrets.
        :param experiment_id: The experiment id to resolve.
        :return: The experiment as dictionary with filled template values.
        """
        experiment = self._mongo.db['experiments'].find_one(
            {'_id': ObjectId(experiment_id)},
            {'container.settings': 1, 'execution.settings': 1}
        )

        experiment = self._fill_experiment_secret_keys(experiment)

        return experiment

    def _fill_experiment_secret_keys(self, experiment):
        """
        Returns the given experiment with filled template keys and values.
        :param experiment: The experiment to complete.
        :return: Returns the given experiment with filled template keys and values.
        :raise TrusteeServiceError: If the trustee service is unavailable or the trustee service could not fulfill all
        requested keys
        """
        experiment_secret_keys = get_experiment_secret_keys(experiment)
        response = self._trustee_client.collect(experiment_secret_keys)
        if response['state'] == 'failed':

            debug_info = response['debugInfo']

            if response.get('inspect'):
                response = self._trustee_client.inspect()
                if response['state'] == 'failed':
                    debug_info = response['debug_info']
                    raise TrusteeServiceError('Trustee service unavailable:{}{}'.format(os.linesep, debug_info))

            experiment_id = str(experiment['_id'])
            raise TrusteeServiceError(
                'Trustee service request failed for experiment "{}":{}{}'.format(experiment_id, os.linesep, debug_info)
            )

        experiment_secrets = response['secrets']
        return fill_experiment_secrets(experiment, experiment_secrets)

    def _fifo(self):
        cursor = self._mongo.db['batches'].aggregate([
            {'$match': {'state': 'registered'}},
            {'$sort': {'registrationTime': 1}},
            {'$project': {'experimentId': 1, 'inputs': 1, 'outputs': 1, 'state': 1}}
        ])
        for b in cursor:
            yield b



import os
from argparse import ArgumentParser
import atexit
from pprint import pprint

import zmq
import pymongo

from cc_core.version import VERSION as CORE_VERSION
from cc_agency.version import VERSION as AGENCY_VERSION
from cc_agency.commons.conf import Conf
from cc_agency.commons.db import Mongo
from cc_agency.commons.secrets import TrusteeClient
from cc_agency.controller.scheduler import Scheduler


DESCRIPTION = 'CC-Agency Controller'


def main():
    print('CC-Agency Version:', AGENCY_VERSION)
    print('CC-Core Version:', CORE_VERSION)

    parser = ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        '-c', '--conf-file', action='store', type=str, metavar='CONF_FILE',
        help='CONF_FILE (yaml) as local path.'
    )
    args = parser.parse_args()

    # Singletons
    conf = Conf(args.conf_file)
    mongo = Mongo(conf)

    # MongoDB indexes
    mongo.db['batches'].create_index([('state', pymongo.ASCENDING)])
    mongo.db['batches'].create_index([('protectedKeysVoided', pymongo.ASCENDING)])
    mongo.db['batches'].create_index([('notificationsSent', pymongo.ASCENDING)])
    mongo.db['batches'].create_index([('experimentId', pymongo.ASCENDING)])
    mongo.db['batches'].create_index([('username', pymongo.ASCENDING)])

    print('MongoDB Indexes:')
    pprint(list(mongo.db['experiments'].list_indexes()) + list(mongo.db['batches'].list_indexes()))

    # Singletons
    trustee_client = TrusteeClient(conf)
    scheduler = Scheduler(conf, mongo, trustee_client)

    # ZeroMQ socket
    bind_socket_path = os.path.expanduser(conf.d['controller']['bind_socket_path'])
    bind_socket_dir, _ = os.path.split(bind_socket_path)

    os.makedirs(bind_socket_dir, exist_ok=True)

    old_umask = os.umask(0o077)
    context = zmq.Context()
    # noinspection PyUnresolvedReferences
    socket = context.socket(zmq.PULL)
    socket.bind('ipc://{}'.format(bind_socket_path))
    os.umask(old_umask)

    atexit.register(socket.close)

    while True:
        data = socket.recv_json()

        if 'destination' not in data:
            continue

        destination = data['destination']
        if destination == 'scheduler':
            scheduler.schedule()



from cc_core.commons.schemas.engines.container import MIN_RAM_LIMIT


conf_schema = {
    'type': 'object',
    'properties': {
        'broker': {
            'type': 'object',
            'properties': {
                'auth': {
                    'type': 'object',
                    'properties': {
                        'num_login_attempts': {'type': 'integer'},
                        'block_for_seconds': {'type': 'integer'},
                        'tokens_valid_for_seconds': {'type': 'integer'}
                    },
                    'additionalProperties': False,
                    'required': ['num_login_attempts', 'block_for_seconds', 'tokens_valid_for_seconds']
                }
            },
            'additionalProperties': False,
            'required': ['auth']
        },
        'controller': {
            'type': 'object',
            'properties': {
                'bind_socket_path': {'type': 'string'},
                'docker': {
                    'type': 'object',
                    'properties': {
                        'nodes': {
                            'type': 'object',
                            'patternProperties': {
                                '^[a-zA-Z0-9_-]+$': {
                                    'type': 'object',
                                    'properties': {
                                        'base_url': {'type': 'string'},
                                        'tls': {
                                            'type': 'object',
                                            'properties': {
                                                'verify': {'type': 'string'},
                                                'client_cert': {
                                                    'type': 'array',
                                                    'items': {'type': 'string'}
                                                },
                                                'assert_hostname': {'type': ['boolean', 'string']}
                                            },
                                            'additionalProperties': True
                                        },
                                        'environment': {
                                            'type': 'object',
                                            'patternProperties': {
                                                '^[a-zA-Z0-9_-]+$': {'type': 'string'}
                                            },
                                            'additionalProperties': False
                                        },
                                        'network': {'type': 'string'},
                                        'hardware': {
                                            'type': 'object',
                                            'properties': {
                                                'gpus': {
                                                    'type': 'array',
                                                    'items': {
                                                        'type': 'object',
                                                        'properties': {
                                                            'id': {'type': 'integer', 'minimum': 0},
                                                            'vram': {'type': 'integer', 'minimum': MIN_RAM_LIMIT},
                                                        },
                                                        'additionalProperties': False,
                                                        'required': ['id']
                                                    }
                                                }
                                            },
                                            'additionalProperties': False
                                        }
                                    },
                                    'required': ['base_url'],
                                    'additionalProperties': False
                                }
                            },
                            'additionalProperties': False
                        },
                        'allow_insecure_capabilities': {'type': 'boolean'}
                    },
                    'additionalProperties': False,
                    'required': ['nodes']
                },
                'notification_hooks': {
                    'type': 'array',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'url': {'type': 'string'},
                            'auth': {
                                'type': 'object',
                                'properties': {
                                    'username': {'type': 'string'},
                                    'password': {'type': 'string'}
                                },
                                'additionalProperties': False,
                                'required': ['username', 'password']
                            }
                        },
                        'additionalProperties': False,
                        'required': ['url']
                    }
                }
            },
            'additionalProperties': False,
            'required': ['bind_socket_path', 'docker']
        },
        'trustee': {
            'type': 'object',
            'properties': {
                'internal_url': {'type': 'string'},
                'username': {'type': 'string'},
                'password': {'type': 'string'}
            },
            'additionalProperties': False,
            'required': ['internal_url', 'username', 'password']
        },
        'mongo': {
            'type': 'object',
            'properties': {
                'host': {'type': 'string'},
                'port': {'type': 'integer'},
                'db': {'type': 'string'},
                'username': {'type': 'string'},
                'password': {'type': 'string'}
            },
            'additionalProperties': False,
            'required': ['db', 'username', 'password']
        }
    },
    'additionalProperties': False,
    'required': ['broker', 'controller', 'trustee', 'mongo']
}
